// - included reference to PhysicalListener  
// - Position, minGain, maxGain, referenceDistance, falloffDistance and
//   falloffFactor included as well as set and get methods
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "bergenServer.h"
#include "bergenSound.h"


// The constructor must be given a pointer to a server object; it will tell
// the server object to add this sound to its list of sounds
bergenSound::bergenSound(bergenServer *server)
	{
	server_ = server;
	handle_ = 0;
	isPlaying_ = false;
	if (server)
		server->addSound(this);
	else
		fprintf(stderr,"ERROR: a sound was created with a NULL server pointer; "
			"this may core dump\n");
	}


// The destructor does an automatic kill before the sound object is removed.
bergenSound::~bergenSound(void)
	{
	kill();
	}


// Process a message from the server
void bergenSound::message(char **msg,int numStrings)
	{
	if (!strcasecmp(msg[0],"stop"))
		isPlaying_ = false;
	}


// Sets the sound's current amplitude; sends a message to the server program 
// to accomplish this.
void bergenSound::setAmplitude(float amp)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d amplitude %.4f", handle_, amp);
		server_->sendMessage(message);
		}
	}


// Sets the sound's current distance; sends a message to the server program 
// to accomplish this.
void bergenSound::setDistance(float dist)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d distance %.4f", handle_, dist);
		server_->sendMessage(message);
		}
	}


// Sends a message to the server program to start playing the sound.
void bergenSound::play(void)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d play", handle_);
		server_->sendMessage(message);
		}
	isPlaying_ = true;
	}


// Sends a message to the server program to stop playing the sound. The next
// time a play command is issued, the sound will start again from the
// beginning.
void bergenSound::stop(void)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d stop", handle_);
		server_->sendMessage(message);
		}
	isPlaying_ = false;
	}


// Sends a message to the server program to pause the sound. The next time a
// play command is issued, the sound will resume from the point at which it
// was paused.
void bergenSound::pause(void)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d pause", handle_);
		server_->sendMessage(message);
		}
	isPlaying_ = false;
	}


// Sends a message to the server program to remove the sound, and tells the
// server object to remove this sound from its list. The sound object should
// not be used after kill() is called; this is meant to be called from the
// destructor, and should not generally be used directly by applications.
void bergenSound::kill(void)
	{
	killRemote();
	if (server_)
		server_->removeSound(this);
	isPlaying_ = false;
	}


// Sends a message to the server program to remove the sound, and tells the
// server object to remove this sound from its list.
void bergenSound::killRemote(void)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d kill", handle_);
		server_->sendMessage(message);
		}
	}


// Not implemented
void bergenSound::createRemote(void)
	{
	}


// Sets the sound's position
void bergenSound::setPosition(float x, float y, float z)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundPosition %f %f %f", handle_,x,y,z);
		server_->sendMessage(message);
		}
	positionX_ = x;
	positionY_ = y;
	positionZ_ = z;
	}
	

// Sets the sound's attenuation type for distance attenuation
void bergenSound::setAttenuationType(AttenuationType type)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundAttenuationType %i", handle_,type);
		server_->sendMessage(message);
		}
	attenuationType_=type;	
	}


// Gets the sound's attenuation type 
AttenuationType bergenSound::attenuationType(void)
	{
	return attenuationType_;
	}


// Sets the sound's min gain 
void bergenSound::setMinGain(float gain)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundMinGain %f", handle_,gain);
		server_->sendMessage(message);
		}
	minGain_=gain;	
	}


// Gets the sound's minGain
float bergenSound::minGain(void)
	{
	return minGain_;
	}


// Sets the sound's max gain 
void bergenSound::setMaxGain(float gain)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundMaxGain %f", handle_,gain);
		server_->sendMessage(message);
		}
	maxGain_=gain;	
	}


// Gets the sound's maxGain
float bergenSound::maxGain(void)
	{
	return maxGain_;
	}


// Sets the sound's reference distance
void bergenSound::setReferenceDistance(float distance)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundReferenceDistance %f", handle_,distance);
		server_->sendMessage(message);
		}
	referenceDistance_=distance;	
	}


// Gets the sound's referenceDistance Value
float bergenSound::referenceDistance(void)
	{
	return referenceDistance_;
	}


// Sets the sound's falloff distance
void bergenSound::setFalloffDistance(float distance)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundFalloffDistance %f", handle_,distance);
		server_->sendMessage(message);
		}
	falloffDistance_=distance;	
	}


// Gets the sound's falloffDitance
float bergenSound::falloffDistance(void)
	{
	return falloffDistance_;
	}


// Sets the sound's falloff factor
void bergenSound::setFalloffFactor(float factor)
	{
	if (server_)
		{
		char message[256];
		sprintf(message,"%d setSoundFalloffFactor %f", handle_,factor);
		server_->sendMessage(message);
		}
	falloffFactor_=factor;	
	}


// Gets the sound's falloffFactor
float bergenSound::falloffFactor(void)
	{
	return falloffFactor_;
	}
