#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "ASound.h"
#include "snerdDB.h"
#include "Listener.h"

#define TEN_EXP_1_20 1.1220184542

int ASound::NextHandle = 1;

// Initialises ASound Object with default values and allocates memory for 
// storing audio data                       
ASound::ASound(snerdDB* database=NULL, int listener=-1, ConfigInformation* config=NULL)
	{
	handle_ = NextHandle++;
	db_ = database;
	if (db_)
		bufferSize_ = db_->bufferSize();
	buffer_ = (short *) malloc(bufferSize_*sizeof(short));
	listenerHandle_ = listener;
	config_ = config;
	givenDistance_ = false;
	distance_ = 0;
	amplitude_ = 1;
	sampleRate_ = 22050;
	reportedX_ = 0.0;
	reportedY_ = 0.0;
	reportedZ_ = 0.0;
	positionX_ = 0.0;
	positionY_ = 0.0;
	positionZ_ = 0.0;
	attenuationType_ = NONE;
	minGain_ = 0.0;
	maxGain_ = 1.0;
	referenceDistance_ = 0.0;
	falloffDistance_ = 10.0;
	falloffFactor_ = 0.01;
	char cmd[256];
	sprintf(cmd,"-1 handle %d",handle_);
	snerdServer::sendMessage(cmd,strlen(cmd)+1,listenerHandle_);
	}

// The memory occupied by the buffer is freed 
ASound::~ASound(void)
	{
	free(buffer_);
	}


// Returns the actual buffer
short* ASound::updateBuffer(void)
	{
	return buffer_;
	}


// returns flag to initialise that object can be removed
int ASound::removable(void)
	{
	return 1;
	}


// Returns 0 to determine that Object is not playing 
int ASound::active(void)
	{
	return 0;
	}


// Sets the amplitude to the new given amplitude
void ASound::setAmplitude(float amp)
	{
	amplitude_ = amp;
	}


// Gets the object's current amplitude
float ASound::amplitude(void)
	{
	return amplitude_;
	}


// Sets the amplitude to the new given amplitude
void ASound::setDistance(float dist)
	{
	distance_ = dist;
	givenDistance_ = true;
	}


// not implemented 
void ASound::play(void)
	{
	printf("play message not implemented\n");
	}


// not implemented 
void ASound::stop(void)
	{
	}


// not implemented 
void ASound::pause(void)
	{
	}


// not implemented 
void ASound::kill(void)
	{
	}


// Executes the right action refereing to the messages coming in 
void ASound::message(char **msg,int numStrings)
	{
#ifdef DEBUG
	printf("ASound got message %s with numStrings %d\n",*msg,numStrings);
#endif
	if (!strcasecmp(msg[0],"play")){
		#ifdef DEBUG
			printf("ASound play...");
		#endif
		play();
		#ifdef DEBUG
			printf("done\n");
		#endif
		}
	else if (!strcasecmp(msg[0],"stop"))
		stop();
	else if (!strcasecmp(msg[0],"pause"))
		pause();
	else if (!strcasecmp(msg[0],"kill"))
		kill();
	else if (!strcasecmp(msg[0],"amplitude"))
		{
		if (numStrings > 1)
			setAmplitude(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"distance"))
		{
		if (numStrings > 1)
			setDistance(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"setSoundPosition"))
		{
		if (numStrings > 3)
			setPosition(atof(msg[1]),atof(msg[2]),atof(msg[3]));
		}
	else if (!strcasecmp(msg[0],"setSoundAttenuationType"))
		{
		if (numStrings > 1)
			switch (atoi(msg[1]))
				{
				case 0 :
					setAttenuationType(NONE);
					break;
				case 1 :
					setAttenuationType(LINEAR_FALLOFF);
					break;
				case 2 :
					setAttenuationType(INVERSE_SQUARE_LAW);
					break;
				case 3 :
					setAttenuationType(LINEAR_FALLOFF_FACTOR);
					break;
				case 4 :
					setAttenuationType(INVERSE_SQUARE_FACTOR);
					break;
				case 5 :
					setAttenuationType(INVERSE_SQUARE_CLAMPED);
					break;
				}
		}
	else if (!strcasecmp(msg[0],"setSoundMinGain"))
		{
		if (numStrings > 1)
			setMinGain(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"setSoundMaxGain"))
		{
		if (numStrings > 1)
			setMaxGain(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"setSoundReferenceDistance"))
		{
		if (numStrings > 1)
			setReferenceDistance(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"setSoundFalloffDistance"))
		{
		if (numStrings > 1)
			setFalloffDistance(atof(msg[1]));
		}
	else if (!strcasecmp(msg[0],"setSoundFalloffFactor"))
		{
		if (numStrings > 1)
			setFalloffFactor(atof(msg[1]));
		}
	else
		printf("sound %d got unrecognized message \"%s\"\n",handle_,msg[0]);
	}

// Sets the object's sampling rate to the given sampling rate
void ASound::setSampleRate(int rate)
	{
	sampleRate_ = rate;
	}


// Sets the object's position to the given position 
void ASound::setPosition(float x, float y, float z)
	{
	reportedX_ = x;
	reportedY_ = y;
	reportedZ_ = z;
	}


// Updates the object's position with given filter
void ASound::updatePosition(void)
	{
	if (active())
		{
		float filter = config_->soundFilter;
		positionX_ = (1 - filter)*positionX_ + filter*reportedX_;
		positionY_ = (1 - filter)*positionY_ + filter*reportedY_;
		positionZ_ = (1 - filter)*positionZ_ + filter*reportedZ_;
		}
	else
		{
		positionX_ = reportedX_;
		positionX_ = reportedX_;
		positionX_ = reportedX_;
		}
	}


// Calculates the attenuation by distance depending on the distance 
// attenuation technique and the source's position 
float ASound::attenuatedGain(void)
	{
	Listener * listener;
	float diffX, diffY, diffZ, distance;
	float attenuatedGain = amplitude_;
	#ifdef DEBUG
		printf("ASound FindListener\n");
	#endif	
        if (givenDistance_)
		distance = distance_;
	else
		{
		listener = db_->findListener(listenerHandle_);
		diffX = listener->positionX() - positionX_;
		diffY = listener->positionY() - positionY_;
		diffZ = listener->positionZ() - positionZ_;
	
		distance_ = sqrt(diffX*diffX+diffY*diffY+diffZ*diffZ);
		distance = distance_;
		}
	#ifdef DEBUG
		printf("ASound Distance : %f\n",distance);
	#endif
	
	
	switch(attenuationType_)
		{
		case LINEAR_FALLOFF:
			if(distance > referenceDistance_)
				attenuatedGain = (amplitude_ - ( (distance-referenceDistance_)/falloffDistance_ ));
			if (attenuatedGain > maxGain_)
				return maxGain_;
			else if (attenuatedGain < minGain_)
				return minGain_;
			else
				return attenuatedGain;
			break;
		case INVERSE_SQUARE_LAW:
			if(distance > referenceDistance_)
				attenuatedGain = attenuatedGain - 20*log10(1 + (TEN_EXP_1_20 - 1.0)*(distance-referenceDistance_)/falloffDistance_);
			if (attenuatedGain > maxGain_)
				return maxGain_;
			else if (attenuatedGain < minGain_)
				return minGain_;
			else
				return attenuatedGain;
			break;
		case LINEAR_FALLOFF_FACTOR:
			if(distance > referenceDistance_)
				attenuatedGain = (amplitude_ - ( (distance-referenceDistance_)*falloffFactor_ ));
			if (attenuatedGain > maxGain_)
				return maxGain_;
			else if (attenuatedGain < minGain_)
				return minGain_;
			else
				return attenuatedGain;
			break;
		case INVERSE_SQUARE_FACTOR:
			attenuatedGain = attenuatedGain - 20*log10(1 + falloffFactor_*(distance-referenceDistance_)/referenceDistance_);
			if (attenuatedGain > maxGain_)
				return maxGain_;
			else if (attenuatedGain < minGain_)
				return minGain_;
			else
				return attenuatedGain;
			break;
		case INVERSE_SQUARE_CLAMPED:
			if(distance > falloffDistance_)
				distance = falloffDistance_;
			else if (distance < referenceDistance_)
				distance = referenceDistance_;
			attenuatedGain = attenuatedGain - 20*log10(1 + falloffFactor_*(distance-referenceDistance_)/referenceDistance_);
			if (attenuatedGain > maxGain_)
				return maxGain_;
			else if (attenuatedGain < minGain_)
				return minGain_;
			else
				return attenuatedGain;
			break;
		default:
			return attenuatedGain;
		}
	}


// Calculates the gain for th given speaker to used for spatialization 
// using a speaker array      
float ASound::spatializedGain(int speakerIndex)
	{
	//spatialization through ambisonics technique
	Listener * listener;
	int numSpeakers=0;
	float gain;
	float sourceX,sourceY,sourceZ,sourceLength;
	double tempangle,angle;
	
	listener = (Listener*)db_->findListener(listenerHandle_);
	if(listener)
		{	
		numSpeakers=listener->numSpeakers();
		SpeakerInformation** speakers=listener->speakers();
		if(numSpeakers>speakerIndex)
			{
			//track listener movement
			sourceX=positionX_-listener->positionX();
			sourceY=positionY_-listener->positionY();
			sourceZ=positionZ_-listener->positionZ();
			
			//calculate unit vector
			sourceLength=(float)sqrt(pow(sourceX,2)+pow(sourceY,2)+
				pow(sourceZ,2));
			
			//if source position is not equal to listener position
			if(sourceLength)
				{
				//build unit source vector
				sourceX/=sourceLength;
				sourceY/=sourceLength;
				sourceZ/=sourceLength;

				switch(config_->mode)
					{
					case STEREO:
						if(numSpeakers>=2)						
							{
							double angleSpeaker1, angleSpeaker2;
							if(sourceZ>0)
								sourceZ*=-1,0;

							angleSpeaker1 = acos((double)(-1*speakers[0]->actualPosZ));
							if(angleSpeaker1>(0.5*M_PI))
								angleSpeaker1-=(2*(angleSpeaker1-(0.5*M_PI)));
							if( (speakers[0]->actualPosX)<0)
								angleSpeaker1*=-1.0;

							angleSpeaker2 = acos((double)(-1*speakers[1]->actualPosZ));
							if(angleSpeaker2>(0.5*M_PI))
								angleSpeaker2-=(2*(angleSpeaker2-(0.5*M_PI)));
							if( (speakers[1]->actualPosX)<0)
								angleSpeaker2*=-1.0;

							angle = acos((double)(-1*sourceZ));
							if(angle>(0.5*M_PI))
								angle-=(2*(angle-(0.5*M_PI)));
							if(sourceX<0)
								angle*=-1.0;

							if(speakerIndex==0)
								{
								if (angle < angleSpeaker2)
									gain=(float)sqrt((angleSpeaker2-angle)/(angleSpeaker2-angleSpeaker1));
								else
									gain=0.0;
								}
							else if(speakerIndex==1)
								{
								if (angle > angleSpeaker1)
									gain=(float)sqrt((angleSpeaker1-angle)/(angleSpeaker1-angleSpeaker2));
								else
									gain=0.0;
								}
							}
						else
							gain=1.0f;
						break;
					case AMBISONIC:
						//calculate 3D angle
						if(numSpeakers==4)						
							{
							tempangle = ((sourceX*speakers[speakerIndex]->actualPosX) + 
							     	     (sourceZ*speakers[speakerIndex]->actualPosZ));
							angle = acos(tempangle);
							//printf("temp angle[%d]: %f\n",speakerIndex,tempangle);
							//mute if angle is bigger than 90 degrees or not (if not decomment if(1) )
							if(fabs(angle)<=M_PI*0.5)
							//if(1)
							//gain = (float)((1.0/numSpeakers) * (1+2*cos(angle)+2*cos(2*angle)));
							gain = (float)((1.0/numSpeakers) * (1+2*cos(angle)));
							else 
								gain = 0.0f;
							}
						else	{
							tempangle = ((sourceX*speakers[speakerIndex]->actualPosX) + 
						        	     (sourceY*speakers[speakerIndex]->actualPosY) +
							     	     (sourceZ*speakers[speakerIndex]->actualPosZ));
							angle = acos(tempangle);
							//mute if angle is bigger than 90 degrees or not (if not decomment if(1) )
							//if(abs(angle)<=M_PI*0.5)
							if(1)
							//gain = (float)((1.0/numSpeakers) * (1+2*cos(angle)+2*cos(2*angle)));
							gain = (float)((1.0/numSpeakers) * (1+2*cos(angle)));
							else 
								gain = 0.0f;
							}
						break;
					case VBAP:
						gain=1.0f;
						break;
					default:
						gain=1.0f;
						break;
					}//endswitch
				}//endif source length
			}//endif speaker amount
		}//endif listener
	if(gain>1.0f)
		gain=1.0f;
	if(gain<0.0f)
		gain=0.0f;
	if (distance_ < 0)
		{
		float delta = 1.0 - gain;
		gain += delta*(-distance_);
		}
	return gain;
	}


// sets the attentuation type for distance attenuation
 void ASound::setAttenuationType(AttenuationType type)
	{
	attenuationType_ = type;
	}


// Sets the sound's min gain
void ASound::setMinGain(float gain)
	{
	minGain_ = gain;
	}
	

// Sets the sound's max gain
void ASound::setMaxGain(float gain)
	{
	maxGain_ = gain;
	}
	

// Sets the sound's reference distance
void ASound::setReferenceDistance(float distance)
	{
	referenceDistance_ = distance;
	}
	

// Sets the sound's falloff distance 
void ASound::setFalloffDistance(float distance)
	{
	falloffDistance_ = distance;
	}
	

// Sets the sound's falloff factor 
void ASound::setFalloffFactor(float factor)
	{
	falloffFactor_ = factor;
	}
