#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include "AudioInterface.h"
#include "Amplitude.h"

#define VOICE_NAVG 10

// Initializes Amplitude Object with initial values
Amplitude::Amplitude(snerdDB* sDB=NULL, int listener=-1, ConfigInformation* config=NULL) :
ASound(sDB,listener,config)
	{
	killed_ = 0;
	playing_ = 0;
	inputBuf_ = NULL;
	bytesPerSample_ = 2;
	gain_ = 0.0;
	channels_ = AudioInterface::inputChannels();
	bufferSize_ = bufferSize_/channels_;
	}


// The input buffer is freed
Amplitude::~Amplitude(void)
	{
	free(inputBuf_);
	}


// Sets back the playing_ flag
void Amplitude::stop(void)
	{
	playing_ = 0;
	}


// Sets the playing_ flag
void Amplitude::play(void)
	{
	playing_ = 1;
	}


// Unsets the playing_ flag 
void Amplitude::pause(void)
	{
	playing_ = 0;
	}


// Sets the killed_ flag
void Amplitude::kill(void)
	{
	killed_ = 1;
	}


// If non of the audio was read before memory for the buffer is allocated.
// Samples to fill one buffer are read from the input file. The buffer is
// returned.  
short * Amplitude::updateBuffer(void)
	{
	updatePosition();
	if (active())
		{
		int errFlag=0;

		if (!inputBuf_)
			inputBuf_ = (short *) malloc(bufferSize_ * sizeof(short));
		errFlag = AudioInterface::readInputPort(inputBuf_, bufferSize_);
		if (errFlag != -1)
			{
			//compute amplitude value
			int max, n, c, i, avg, gavg;
			short *iptr, *sptr;
			int ambient = 0;

			n = bufferSize_/2;	
			c = n;
			sptr = (short *) inputBuf_;
			iptr = sptr;
			
			avg = 0;
			while(c--)
				{
				avg += *iptr++;
				}
			gavg = avg/n;
		
			max = 0;
			c = n;
			while(c)
				{
				avg = 0;
				for(i=0; (i<VOICE_NAVG) && c; i++)
					{
					*sptr -= gavg;
					if(*sptr>0)
						avg += *sptr++;
					else
						avg -= *sptr++;
					c--;
					}
				avg = avg/VOICE_NAVG;
				if(avg>max)
					max = avg;
				}
		
			if(max<ambient)
				ambient = max;
			ambient += 10;
			gain_ = ((max-ambient)/10000.0)*attenuatedGain();
			}
		}

	return NULL;
	}


// A sample file can be removed it was killed before
int Amplitude::removable(void)
	{
	return killed_;
	}


// Returns the value of the playing flag to determine if object is playing 
int Amplitude::active(void)
	{
	return playing_;
	}


// Executes the right action refering to the messages coming in
void Amplitude::message(char **msg, int numStrings)
	{
	if (!strcasecmp(msg[0],"getamplitude"))
		{
		char cmd[256];
		sprintf(cmd,"%d amplitude %f",handle_,gain_);
		snerdServer::sendMessage(cmd,strlen(cmd)+1,listenerHandle_);
		}
	else
		ASound::message(msg, numStrings);
	}
