#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "snerdServer.h"
#include "Listener.h"

int Listener::NextHandle = 1;

// Listener Constructor sets the initial position to 0,0,0 sets the 
// killed flag and increases the handle counter as well as it sets the initial
// spealer positions for each sepaker 
Listener::Listener(ConfigInformation* config, char* addr)
	{
	handle_ = NextHandle++;
	killed_ = 0;
	reportedX_ = 0.0;
	reportedY_ = 0.0;
	reportedZ_ = 0.0;
	positionX_ = 0.0;
	positionY_ = 0.0;
	positionZ_ = 0.0;
	config_ = config;
	strcpy(address_,addr);
	for(int i=0;i<config_->numSpeakers;i++)
		{
		config_->speakers[i]->initialPosX=config_->speakers[i]->actualPosX;
		config_->speakers[i]->initialPosY=config_->speakers[i]->actualPosY;
		config_->speakers[i]->initialPosZ=config_->speakers[i]->actualPosZ;
		}
	updateSpeakers();
	char cmd[256];
	sprintf(cmd,"-1 handle %d",handle_);
	snerdServer::sendMessage(cmd,strlen(cmd)+1,handle_);
	}	


// The Listener destructor 
Listener::~Listener(void)
	{	
	}	


// Sets the destination address of the listener for return messages
void Listener::setAddress(char* addr)
	{
	strcpy(address_,addr);
	}

// Sets the lsitener position to the new given position and calls
// updateSpeakers to calculate the actual speaker positions 
void Listener::setPosition(float x, float y, float z)
	{
	reportedX_ = x;
	reportedY_ = y;
	reportedZ_ = z;
	}
	

// Sets the lsitener position to the new given position and calls
// updateSpeakers to calculate the actual speaker positions 
void Listener::updatePosition(void)
	{
	float filter = config_->listenerFilter;
	positionX_ = (1 - filter)*positionX_ + filter*reportedX_;
	positionY_ = (1 - filter)*positionY_ + filter*reportedY_;
	positionZ_ = (1 - filter)*positionZ_ + filter*reportedZ_;
	updateSpeakers();
	}
	

// Sets the speaker position and calls update speaker to get unit vector
void Listener::setSpeakerPosition(int speakerIndex, float x, float y, float z)
	{
	if(config_->speakers && (config_->numSpeakers>(speakerIndex+1)))
		{
		config_->speakers[speakerIndex]->initialPosX=x;
		config_->speakers[speakerIndex]->initialPosY=y;
		config_->speakers[speakerIndex]->initialPosZ=z;
		}
	}
	

// Updates the actual speaker position, taking into acount the initial speaker
// position and the listener position
void Listener::updateSpeakers(void)
	{
	int speakerIndex;
	float vectorLength;
	
	for(speakerIndex=0;speakerIndex<config_->numSpeakers;speakerIndex++)
		{
		config_->speakers[speakerIndex]->actualPosX=config_->speakers[speakerIndex]->initialPosX-positionX_;
		config_->speakers[speakerIndex]->actualPosY=config_->speakers[speakerIndex]->initialPosY-positionY_;
		config_->speakers[speakerIndex]->actualPosZ=config_->speakers[speakerIndex]->initialPosZ-positionZ_;


		if (config_->numSpeakers <= 4)
			{
			vectorLength=(float)sqrt(pow(config_->speakers[speakerIndex]->actualPosX,2)
									+pow(config_->speakers[speakerIndex]->actualPosZ,2));
			if(vectorLength)
				{
				config_->speakers[speakerIndex]->actualPosX/=vectorLength;
				config_->speakers[speakerIndex]->actualPosY=0.0;
				config_->speakers[speakerIndex]->actualPosZ/=vectorLength;
				}
			}
		else
			{
			vectorLength=(float)sqrt(pow(config_->speakers[speakerIndex]->actualPosX,2)
									+pow(config_->speakers[speakerIndex]->actualPosY,2)
									+pow(config_->speakers[speakerIndex]->actualPosZ,2));
			if(vectorLength)
				{
				config_->speakers[speakerIndex]->actualPosX/=vectorLength;
				config_->speakers[speakerIndex]->actualPosY/=vectorLength;
				config_->speakers[speakerIndex]->actualPosZ/=vectorLength;
				}
			}
		}
	}


// Parses the message and calls appropiate function 
void Listener::message(char **msg,int numStrings)
	{
	if (!strcasecmp(msg[0],"setListenerPosition"))
		{
		if (numStrings > 3)
			setPosition(atof(msg[1]),atof(msg[2]),atof(msg[3]));
		}

	else if (!strcasecmp(msg[0],"killListener"))
		{
		kill();
		}
	else
		printf("listener %d got unrecognized message \"%s\"\n",handle_,msg[0]);
	}


// Sets the Listeners killed_ flag to mark it for remove 
void Listener::kill(void)
	{
	killed_ = 1;
	}


// Checks and returs the killed_ flag
int Listener::removable(void)
	{
	return killed_;
	}

