extern "C"
{
  #include <htmsocket.h>
}
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include "bergenUDPSocket.h"
#include "OSCServer.h"

#ifdef __sgi
#define socklen_t int 
#endif

#define WAIT_TICKS 5000

// Setting the default values
bergenUDPSocket * OSCServer::rcvSocket_ = NULL;
bergenUDPSocket * OSCServer::sndSocket_ = NULL;
bool OSCServer::looping_ = false;
ConfigInformation* OSCServer::config_ = NULL;
int OSCServer::socket_osc_ = -1;
void * OSCServer::socket_htm_osc_ = NULL;
OSCbuf OSCServer::buf[1];
char OSCServer::bufferForOSCbuf[500];

// OSCServer Constructor initialises OSCServer object with given 
// buffer size, sample rate and the network connection socket 
void OSCServer::init(ConfigInformation* config)
	{
	config_ = config;
	rcvSocket_ = new bergenUDPSocket(config_->udpPort);
	rcvSocket_->setBlocking(0);
	int port_osc_ = config_->oscPort;
	socket_htm_osc_ = OpenHTMSocket(config_->oscServer, port_osc_);
	if (!socket_htm_osc_)
		{
		printf("OSCServer: Failed to open OSC HTM socket\n");
		return;
		}    
	OSC_initBuffer(buf, 500, bufferForOSCbuf);
	socket_osc_ = socket(AF_INET, SOCK_DGRAM, 0);
	if (socket_osc_ == -1)
		{
		printf("OSCServer: Failed to open OSC UDP socket\n");
		return;
		}
	struct sockaddr_in local_addr;
	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons(port_osc_+1);
	local_addr.sin_addr.s_addr = INADDR_ANY;
	memset(&(local_addr.sin_zero), '\0', 8); //zero the rest of the struct
 	if(bind(socket_osc_, (struct sockaddr *) &local_addr, sizeof(struct sockaddr)) == -1)
		{
		printf("OSCServer: Failed to bind OSC UDP socket\n");
		return;
		}
	fcntl(socket_osc_, F_SETFL, FNDELAY); //non-blocking
	char spatial[64] = "spatialization ";
	if (config_->mode == BINAURAL)
		strcat(spatial,"binaural");
	else if (config_->mode == STEREO)
		strcat(spatial,"stereo");
	else if (config_->mode == AMBISONIC)
		strcat(spatial,"ambisonic");
	else if (config_->mode == VBAP)
		strcat(spatial,"vbap");
	else
		strcat(spatial,"none");
	sendOSCString("snerd",spatial);
	for (int i=0;i<config_->numSpeakers;i++)
		{
		char message[256];
		float pos[3];
		pos[0] = config_->speakers[i]->actualPosX;
		pos[1] = config_->speakers[i]->actualPosY;
		pos[2] = config_->speakers[i]->actualPosZ;
		sendOSCStringVec3("snerd","speaker",pos);
		}
	}


// Once called the mainLoop functions only ends when the snerd server attribute
// "looping_" is disabled.
// With every loop new arriving messages are processed, the sound objects are
// updated and the updated and mixed audiobuffer is written to the audiodevice
void OSCServer::mainLoop(void)
	{
	looping_ = true;
	while (looping_)
		processMessages();
	if (socket_htm_osc_)
		CloseHTMSocket(socket_htm_osc_);
	delete rcvSocket_;
	delete sndSocket_;
	}


// If a message was received over the socket, the message is send to 
// splitMessage for getting messages isolated. The isolated maessages are
// parsed by sending them to parseMessage if they are no sound related ones.
// Sound messages are forwarded to the according sound object
void OSCServer::processMessages(void)
	{
	char message[1024], *msgStrings[32];
	int numStrings;
	while (rcvSocket_->receive(message,sizeof(message)) > 0)
		{
		numStrings = splitMessage(message, msgStrings, 32);
		if (numStrings <= 0)
			continue;
		if (!strcasecmp(msgStrings[1],"kill"))
			looping_ = 0;
		else if (!strcasecmp(msgStrings[1],"ping"))
			{
			char* reply = "-1 pong";
			sndSocket_ = new bergenUDPSocket(0);
			sndSocket_->setDestination(rcvSocket_->lastFromAddress(),config_->udpPort+1);
			sndSocket_->setBlocking(0);
			sendMessage(reply,strlen(reply)+1);
			}
		else if (!strcasecmp(msgStrings[1],"new") && !strcasecmp(msgStrings[2],"Listener"))
			{
			char cmd[256];
			sprintf(cmd,"-1 handle 1");
			sendMessage(cmd,strlen(cmd)+1);
			}
		else if (!strcasecmp(msgStrings[1],"SetListenerPosition"))
			continue;
		else if (!strcasecmp(msgStrings[1],"new"))
			{
			sendOSCString("snerd",message);
			int handle;
			handle = receiveOSCHandle();
			char cmd[256];
			sprintf(cmd,"-1 handle %d",handle);
			sendMessage(cmd,strlen(cmd)+1);
			}
		else if (!strcasecmp(msgStrings[1],"getamplitude"))
			{
			sendOSCString("snerd",message);
			// not currently implemented
			//float value;
			//while(!receiveOSCFloat(value))
			//	;
			//returnFloat(value);
			}
		else
			sendOSCString("snerd",message);
		while (numStrings)
			{
			numStrings--;
			free(msgStrings[numStrings]);
			}
		}
	if (socket_osc_ >= 0)
		{
		fd_set read_fds;
		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		FD_ZERO(&read_fds);
		FD_SET(socket_osc_, &read_fds);
		select(socket_osc_+1, &read_fds, NULL, NULL, &tv);
		if(FD_ISSET(socket_osc_, &read_fds))
			{
			int addr_len;
			int numbytes;
			char buf[500];
			addr_len = sizeof(struct sockaddr);
			struct sockaddr_in receive_addr;
			while((numbytes = recvfrom(socket_osc_, buf, 500, 0, (struct sockaddr *)&receive_addr, (socklen_t*)&addr_len)) > 0)
				{
				numStrings = parseOSCPacket(buf, numbytes, msgStrings, 32, 0) ;
				if (numStrings > 0)
					{
					char cmd[256];
					strcpy(cmd,"");
					for (int i=0; i<numStrings; i++)
						{
						if (i > 0)
							strcat(cmd," ");
						strcat(cmd,msgStrings[i]);
						}
					sendMessage(cmd,strlen(cmd)+1);
					}
				while (numStrings)
					{
					numStrings--;
					free(msgStrings[numStrings]);
					}
				}
			}
		}
	//sleep to prevent spinlocking
	struct timespec sleeptime;
	sleeptime.tv_sec = 0;
	sleeptime.tv_nsec = 1000000;
	nanosleep(&sleeptime,NULL);
	}


// The one message string is split into isolated messages. Each message is
// either seperated by \t \n or ' '. The amount of messages found is returned
int OSCServer::splitMessage(char *message,char **strings,int max)
	{
	int i=0, start, len, count=0;
	while ((message[i]) && (count < max))
		{
		i += strspn(message+i, " \t\n");
		if (!message[i])
			break;
		start = i;
		len = strcspn(message+start, " \t\n");
		strings[count] = (char *) malloc(len+1);
		strncpy(strings[count],message+start,len);
		strings[count][len] = '\0';
		count++;
		i += len;
		}
	return count;
	}


// Sends back the message the client just sent to the server
void OSCServer::sendMessage(void *msg,int len)
	{
	sndSocket_->send(msg,len);
	}


void OSCServer::sendOSCString(char* r,char* s)
	{
	if (socket_htm_osc_)
		{
		int returnVal;
		returnVal = OSC_writeAddressAndTypes(buf, r, ",s");
		if (returnVal) 
			printf("OSCServer: Bad return value from OSC_writeAddress\n");
		else
			{
			returnVal = OSC_writeStringArg(buf, s); 
			sendOSCBuffer();
			OSC_resetBuffer(buf);
			}
		}
	}


void OSCServer::sendOSCStringVec3(char* r,char* s,float* v)
	{
	if (socket_htm_osc_)
		{
		int returnVal;
		returnVal = OSC_writeAddressAndTypes(buf, r, ",sfff");
		if (returnVal) 
			printf("OSCServer: Bad return value from OSC_writeAddress\n");
		else
			{
			returnVal = OSC_writeStringArg(buf, s);
			returnVal = OSC_writeFloatArgs(buf, 3, v);
			sendOSCBuffer();
			OSC_resetBuffer(buf);
			}
		}
	}


void OSCServer::sendOSCBuffer(void) 
	{
	if (OSC_isBufferEmpty(buf)) 
		return;
	if (!OSC_isBufferDone(buf)) 
		{
		printf("OSCServer: sendBuffer() called but buffer not ready!\n");
		return;
		}
	if (socket_htm_osc_ && !SendHTMSocket(socket_htm_osc_, OSC_packetSize(buf), OSC_getPacket(buf))) 
		{
		printf("OSCServer: Couldn't send out socket\n");
		CloseHTMSocket(socket_htm_osc_);
		socket_htm_osc_ = NULL;
		}
	}


// Wait for the next message from the OSC server
int OSCServer::receiveOSCHandle(void)
	{
	if (socket_osc_ >= 0)
		{
		char *msgStrings[32];
		int numStrings;
		int wait;
		int handle;
		for (wait=0; wait < WAIT_TICKS; wait++)
			{
			fd_set read_fds;
			struct timeval tv;
			tv.tv_sec = 0;
			tv.tv_usec = 0;
			FD_ZERO(&read_fds);
			FD_SET(socket_osc_, &read_fds);
			select(socket_osc_+1, &read_fds, NULL, NULL, &tv);
			if(FD_ISSET(socket_osc_, &read_fds))
				{
				int addr_len;
				int numbytes;
				char buf[500];
				addr_len = sizeof(struct sockaddr);
				struct sockaddr_in receive_addr;
				if((numbytes = recvfrom(socket_osc_, buf, 500, 0, (struct sockaddr *)&receive_addr, (socklen_t*)&addr_len)) > 0)
					{
					bool exit = false;
					numStrings = parseOSCPacket(buf, numbytes, msgStrings, 32, 0) ;
					if (numStrings == 3 &&	 !strcasecmp(msgStrings[1],"handle"))
						{
						handle = atoi(msgStrings[2]);
						exit = true;
						}
					while (numStrings)
						{
						numStrings--;
						free(msgStrings[numStrings]);
						}
					if (exit)
						return handle;
					}
				usleep(10000);
				}
			}
		fprintf(stderr,"WARNING: OSCServer failed to receive a handle from (%s)\n",config_->oscServer);
		}
	return 0;
	}


int OSCServer::parseOSCPacket(char *buf, int n, char **strings, int max, int pos) {
    int size, messageLen, i;
    char *messageName;
    char *args;

    if ((n%4) != 0) {
	printf("OSC packet size (%d) not a multiple of 4 bytes: dropping\n",
		 n);
	return 0;
    }

    if ((n >= 8) && (strncmp(buf, "#bundle", 8) == 0)) {
	/* This is a bundle message. */

	if (n < 16) {
	    printf("OSC Bundle message too small (%d bytes) for time tag\n", n);
	    return 0;
	}

	i = 16; /* Skip "#group\0" and time tag */
	while(i<n) {
	    size = ntohl(*((int *) (buf + i)));
	    if ((size % 4) != 0) {
		printf("OSC size count %d in bundle bad (not a multiple of 4)\n", size);
		return 0;
	    }
	    if ((size + i + 4) > n) {
		printf("OSC size count %d in bundle bad (only %d bytes left in entire bundle)\n",
			 size, n-i-4);
		return 0;	
	    }
	    
	    /* Recursively handle element of bundle */
	    return parseOSCPacket(buf+i+4, size, strings, max, pos);
	    i += 4 + size;
	}
	if (i != n) {
	    printf("This can't happen");
	}
    } else {
	/* This is not a bundle message */

	messageName = buf;
	args = dataAfterAlignedString(messageName, buf+n);
	if (args == 0) {
	    printf("OSC Dropping entire message\n");
	    return 0;
	}
	messageLen = args-messageName;
	return smessage(messageName, (void *)args, n-messageLen, strings, max, pos);
    }
}

#define SMALLEST_POSITIVE_FLOAT 0.000001f

int OSCServer::smessage(char* addr, void *v, int n, char **strings, int max, int pos) {
    char *chars = (char*)v;
    strings[pos] = (char *)malloc(strlen(addr)+1);
    strcpy(strings[pos],addr);
    pos++;

    if (n != 0) {
	if (chars[0] == ',') {
	    if (chars[1] != ',') {
		/* This message begins with a type-tag string */
		return parseTypeTaggedArgs(v, n, strings, max, pos);
	    } else {
		/* Double comma means an escaped real comma, not a type string */
		return parseHeuristicallyTypeGuessedArgs(v, n, 1, strings, max, pos);
	    }
	} else {
	    return parseHeuristicallyTypeGuessedArgs(v, n, 0, strings, max, pos);
	}
    }
    return pos;
}

int OSCServer::parseTypeTaggedArgs(void *v, int n, char **strings, int max, int pos) { 
    char *typeTags, *thisType;
    char *p;

    typeTags = (char*)v;

    if (!isNiceString(typeTags, typeTags+n)) {
	/* No null-termination, so maybe it wasn't a type tag
	   string after all */
	return parseHeuristicallyTypeGuessedArgs(v, n, 0, strings, max, pos);
    }

    p = dataAfterAlignedString(typeTags, typeTags+n);

    char argument[256];
    for (thisType = typeTags + 1; *thisType != 0; ++thisType) {
	switch (*thisType) {
	    case 'i': case 'r': case 'm': case 'c':
	    sprintf(argument,"%d", ntohl(*((int *) p)));
	    p += 4;
	    break;

	    case 'f': {
		int i = ntohl(*((int *) p));
		float *floatp = ((float *) (&i));
		sprintf(argument,"%f", *floatp);
		p += 4;
	    }
	    break;

	    case 'h': case 't':
	    sprintf(argument,"[A 64-bit int]");
	    p += 8;
	    break;

	    case 'd':
	    sprintf(argument,"[A 64-bit float]");
	    p += 8;
	    break;

	    case 's': case 'S':
	    if (!isNiceString(p, typeTags+n)) {
		printf("Type tag said this arg is a string but it's not!\n");
		return pos;
	    } else {
		sprintf(argument,"%s", p);
		p = dataAfterAlignedString(p, typeTags+n);
	    }
	    break;

	    case 'T': sprintf(argument,"[True]"); break;
	    case 'F': sprintf(argument,"[False]"); break;
	    case 'N': sprintf(argument,"[Nil]"); break;
	    case 'I': sprintf(argument,"[Infinitum]"); break;

	    default:
	    sprintf(argument,"[Unrecognized type tag %c]", *thisType);
	    return pos;
	}
    strings[pos] = (char *)malloc(strlen(argument)+1);
    strcpy(strings[pos],argument);
    pos++;
    if (pos == max)
        return pos-1;
    }
    return pos;
}


int OSCServer::parseHeuristicallyTypeGuessedArgs(void *v, int n, int skipComma, char **strings, int max, int pos) {
    int i, thisi;
    float thisf;
    int *ints;
    char *chars;
    char *string, *nextString;
	

    /* Go through the arguments 32 bits at a time */
    ints = (int*)v;
    chars = (char*)v;

    char argument[256];
    for (i = 0; i<n/4; ) {
	string = &chars[i*4];
	thisi = ntohl(ints[i]);
	/* Reinterpret the (potentially byte-reversed) thisi as a float */
	thisf = *(((float *) (&thisi)));

	if  (thisi >= -1000 && thisi <= 1000000) {
	    sprintf(argument,"%d", thisi);
	    i++;
	} else if (thisf >= -1000.f && thisf <= 1000000.f &&
		   (thisf <=0.0f || thisf >= SMALLEST_POSITIVE_FLOAT)) {
	    sprintf(argument,"%f",  thisf);
	    i++;
	} else if (isNiceString(string, chars+n)) {
	    nextString = dataAfterAlignedString(string, chars+n);
	    sprintf(argument,"%s", (i == 0 && skipComma) ? string +1 : string);
	    i += (nextString-string) / 4;
	} else {
	    sprintf(argument,"0x%x", ints[i]);
	    i++;
	}
        strings[pos] = (char *)malloc(strlen(argument)+1);
        strcpy(strings[pos],argument);
        pos++;
        if (pos == max)
            return pos-1;
    }
    return pos;
}


#define STRING_ALIGN_PAD 4

char* OSCServer::dataAfterAlignedString(char *string, char *boundary) 
{
    /* The argument is a block of data beginning with a string.  The
       string has (presumably) been padded with extra null characters
       so that the overall length is a multiple of STRING_ALIGN_PAD
       bytes.  Return a pointer to the next byte after the null
       byte(s).  The boundary argument points to the character after
       the last valid character in the buffer---if the string hasn't
       ended by there, something's wrong.

       If the data looks wrong, return 0, and set htm_error_string */

    int i;

    if ((boundary - string) %4 != 0) {
	printf("OSC Internal error: DataAfterAlignedString: bad boundary\n");
	return 0;
    }

    for (i = 0; string[i] != '\0'; i++) {
	if (string + i >= boundary) {
	    printf("OSC DataAfterAlignedString: Unreasonably long string\n");
	    return 0;
	}
    }

    /* Now string[i] is the first null character */
    i++;

    for (; (i % STRING_ALIGN_PAD) != 0; i++) {
	if (string + i >= boundary) {
	    printf("OSC DataAfterAlignedString: Unreasonably long string\n");
	    return 0;
	}
	if (string[i] != '\0') {
	    printf("OSC DataAfterAlignedString: Incorrectly padded string\n");
	    return 0;
	}
    }

    return string+i;
}

#include <ctype.h>
bool OSCServer::isNiceString(char *string, char *boundary) 
{
    /* Arguments same as DataAfterAlignedString().  Is the given "string"
       really a string?  I.e., is it a sequence of isprint() characters
       terminated with 1-4 null characters to align on a 4-byte boundary? */

    int i;

    if ((boundary - string) %4 != 0) {
	printf("OSC Internal error: IsNiceString: bad boundary\n");
	return 0;
    }

    for (i = 0; string[i] != '\0'; i++) {
	if (!isprint(string[i])) return false;
	if (string + i >= boundary) return false;
    }

    /* If we made it this far, it's a null-terminated sequence of printing characters 
       in the given boundary.  Now we just make sure it's null padded... */

    /* Now string[i] is the first null character */
    i++;
    for (; (i % STRING_ALIGN_PAD) != 0; i++) {
	if (string[i] != '\0') return false;
    }

    return true;
}


