#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include "AudioInterface.h"
#include "snerdDB.h"
#include "RecordFile.h"


RecordFile::RecordFile(char *filename, snerdDB* sDB=NULL, int listener=-1, ConfigInformation* config=NULL) :
ASound(sDB,listener,config)
	{
	killed_ = 0;
	char fullpath[256];
	db_->findFile(filename,fullpath);
	filename_ = strdup(fullpath);
	recording_ = 0;
	inputBuf_ = NULL;
	initializeAFFile();
	}


// Initializes recorded file format
void RecordFile::initializeAFFile(void)
	{
#ifdef AUDIOFILE_LIBRARY

	bytesPerSample_ = 2; 					//for sample width 16

	//Set up file format for the recording
	filesetup_    = afNewFileSetup();

	afInitFileFormat(filesetup_, AF_FILE_AIFF);

	sampleRate_ = AudioInterface::inputRate();
	channels_ = AudioInterface::inputChannels();
	bufferSize_ = bufferSize_/channels_;
	afInitChannels(filesetup_, AF_DEFAULT_TRACK, 1);
	afInitRate(filesetup_, AF_DEFAULT_TRACK, sampleRate_);
 
	long filesampwidth = 16;    	// file sample format 
	afInitSampleFormat(filesetup_, AF_DEFAULT_TRACK, AF_SAMPFMT_TWOSCOMP, filesampwidth); // in bits 

	afFile_   = NULL;
#else
	afFile_ = AF_NULL_FILEHANDLE;
	fprintf(stderr,"No SGI audiofile library support - cannot handle file \"%s\"\n",
		filename_);
#endif
	}


// The file handle is closed 
RecordFile::~RecordFile(void)
	{
	free(inputBuf_);
#ifdef AUDIOFILE_LIBRARY
	if (afFile_ != AF_NULL_FILEHANDLE)
		afCloseFile(afFile_);
#endif
	}


// Returns the filename of the audio file that is associated with the
// sample file object   
char * RecordFile::filename(void)
	{
	return filename_;
	}


// Sets the recording_ flag and opens the file handle                                                    *
void RecordFile::record(void)
	{
#ifdef AUDIOFILE_LIBRARY
	// Open a new file for recording
	afFile_ = afOpenFile(filename_, "w", filesetup_);
  
	if (afFile_ == AF_NULL_FILEHANDLE) {
		fprintf(stderr,"ERROR: failed to open sample file \"%s\"\n",filename_);
		perror("afOpenFile");
	}

	recording_ = 1;
	free(inputBuf_);
	inputBuf_ = NULL;
#else
	afFile_ = AF_NULL_FILEHANDLE;
	fprintf(stderr,"No SGI audiofile library support - cannot handle file \"%s\"\n",
		filename_);
#endif
	}


// Sets back the recording_ flag and closes the file handle 
void RecordFile::stop(void)
	{
	recording_ = 0;
#ifdef AUDIOFILE_LIBRARY
	if (afFile_ != AF_NULL_FILEHANDLE)
		afCloseFile(afFile_);
	afFile_ = AF_NULL_FILEHANDLE;
#endif
	}


// Not implemented with recording
void RecordFile::play(void)
	{
	printf("play not implemented with RecordFile\n");
	}


// Unsets the recording_ flag
void RecordFile::pause(void)
	{
	recording_ = 0;
	}


// Sets the killed_ flag and closes the associated audio file
void RecordFile::kill(void)
	{
	killed_ = 1;
#ifdef AUDIOFILE_LIBRARY
	if (afFile_ != AF_NULL_FILEHANDLE)
		afCloseFile(afFile_);
	afFile_ = AF_NULL_FILEHANDLE;
#endif
	}


// If none of the audio was read before memory for the buffer is allocated.
// Samples to fill one buffer are read from the input file. The buffer is 
// returned.
short * RecordFile::updateBuffer(void)
	{
	updatePosition();
	if (active())
		{
		int errFlag=0;

		if (!inputBuf_)
			inputBuf_ = (short *) malloc(bufferSize_ * sizeof(short));
		errFlag = AudioInterface::readInputPort(inputBuf_, bufferSize_);
		float gain;
		gain = attenuatedGain();
		for (int i=0; i<bufferSize_; i++)
			inputBuf_[i] = inputBuf_[i]*gain;
		if (afFile_ && errFlag != -1)
			afWriteFrames(afFile_, AF_DEFAULT_TRACK, inputBuf_, bufferSize_);
		}

	return NULL;
	}


// A sample file can be removed it was killed before
int RecordFile::removable(void)
	{
	return killed_;
	}


// Returns the value of the playing flag to determine if WhiteNoise Objekt is
// playing
int RecordFile::active(void)
	{
#ifdef AUDIOFILE_LIBRARY
	return ((recording_) && (afFile_ != AF_NULL_FILEHANDLE));
#else
	return 0;
#endif
	}


// Executes the right action refering to the messages coming in
void RecordFile::message(char **msg, int numStrings)
	{
	if (!strcasecmp(msg[0],"record"))
		record();
	else
		ASound::message(msg, numStrings);
	}
