#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "Tone.h"

#include <stdio.h>


// Initializes ToneFile Object with initial values 
Tone::Tone(snerdDB* sDB=NULL, int listener=-1, ConfigInformation* config=NULL)
: ASound(sDB,listener,config)
	{
	frequency_ = 1000;
	playing_ = 0;
	killed_ = 0;
	count_ = 0;
	}

// Calls kill routine in order to set killed_ flag to be marked for 
// removable
Tone::~Tone(void)
	{
	kill();
	}


// Samples of given frequency are generated 
short * Tone::updateBuffer(void)
	{
#ifdef DEBUG
	printf("Spatialized Tone updateBuffer\n");
#endif
	updatePosition();
	if (playing_)
		{
		int i,k;
		k=0;
		switch (config_->mode)
			{
			case STEREO:
				for (i=0; (i < (bufferSize_/2)); i++, count_++)
					{
					float value = sinf(count_ * frequency_ * 2.0f * M_PI / sampleRate_)
							* 32768.0f * attenuatedGain();
					buffer_[k++] = (short) (value * spatializedGain(0));
					buffer_[k++] = (short) (value * spatializedGain(1));		
					}
				break;
			case AMBISONIC:	
				for (i=0; (i < (bufferSize_/4)); i++, count_++)
					{
					float value = sinf(count_ * frequency_ * 2.0f * M_PI / sampleRate_)
							* 32768.0f * attenuatedGain();
					buffer_[k++] = (short) (value * spatializedGain(0));
					buffer_[k++] = (short) (value * spatializedGain(1));		
					buffer_[k++] = (short) (value * spatializedGain(2));
					buffer_[k++] = (short) (value * spatializedGain(3));		
					}
				break;
			default:
				for (i=0; i < bufferSize_; i++, count_++)
					buffer_[i] = (short) (sinf(count_ * frequency_ * 2.0f * M_PI / sampleRate_)
							* 32768.0f * attenuatedGain());
				break;
			}
		}
	return buffer_;
	}


// A sample file can be removed it was killed before
int Tone::removable(void)
	{
	return killed_;
	}


// Returns the value of the playing flag to determine if WhiteNoise object is playing
int Tone::active(void)
	{
	return playing_;
	}


// Sets the playing_ flag for the sample file 
void Tone::play(void)
	{
	#ifdef DEBUG
		printf("Tone play routine\n");
	#endif
	playing_ = 1;
	}


// Sets back the playing_ flag and sets the file reading position to the
// start of the audiofile
void Tone::stop(void)
	{
	playing_ = 0;
	}


// Unsets the playing_ flag
void Tone::pause(void)
	{
	playing_ = 0;
	}


// Sets the killed_ flag and unsets the playing_ flag through calling the
// stop routine
void Tone::kill(void)
	{
	stop();
	killed_ = 1;
	}


// Executes the right action refereing to the messages coming in
void Tone::message(char **msg,int numStrings)
	{
#ifdef DEBUG
	printf("Tone got message %s\n",*msg);
#endif
	if (!strcasecmp(msg[0],"setfrequency"))
		{
		if (numStrings > 1)
			setFrequency(atoi(msg[1]));
		}
	else
#ifdef DEBUG
	printf("Forwarding message to ASpatializedSound\n");
#endif
		ASound::message(msg,numStrings);
	}


// Sets the frequency to the given value 
void Tone::setFrequency(int freq)
	{
	frequency_ = freq;
	}
