#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "bergenUDPSocket.h"
#include "snerdConfig.h"
#include "snerdServer.h"
#include "Listener.h"
#include "AudioInterface.h"

#ifdef OSC_DIR
#include "OSCServer.h"
#endif

void loadConfigFile(ConfigInformation* config,char* file)
	{
	FILE *fp=NULL;
	fp = fopen(file,"r");
	if (!fp)
		return;
	char line[512];
	int curSpeakers=0;
	printf("SNERD: Reading configuration file %s\n",file);
	while (fgets(line,sizeof(line),fp))
		{
		char token[32];
		sscanf(line,"%s",token);
		if (!strcmp(token,"speaker"))
			{
			float positionX_,positionY_,positionZ_;
			sscanf(line,"%s%f%f%f",token,&positionX_,&positionY_,&positionZ_);
			if (curSpeakers >= config->numSpeakers)
				config->speakers[curSpeakers] = (SpeakerInformation*)malloc(sizeof(SpeakerInformation));
			config->speakers[curSpeakers]->actualPosX = positionX_;
			config->speakers[curSpeakers]->actualPosY = positionY_;
			config->speakers[curSpeakers]->actualPosZ = positionZ_;
			curSpeakers++;
			}
		else if (!strcmp(token,"spatialization"))
			{
			char technique[32];
			sscanf(line,"%s%s",token,technique);
			if (!strcmp(technique,"mono") || !strcmp(token,"none"))
				config->mode = NO_SPATIALIZATION;
			else if (!strcmp(technique,"binaural"))
				config->mode = BINAURAL;
			else if (!strcmp(technique,"stereo"))
				config->mode = STEREO;
			else if (!strcmp(technique,"quad") || !strcmp(token,"ambisonic"))
				config->mode = AMBISONIC;
			else if (!strcmp(technique,"vbap"))
				config->mode = VBAP;
			}
		else if (!strcmp(token,"srate"))
			{
			int val;
			sscanf(line,"%s%d",token,&val);
			config->sampleRate = val;
			}
		else if (!strcmp(token,"udpport"))
			{
			int val;
			sscanf(line,"%s%d",token,&val);
			config->udpPort = val;
			}
		else if (!strcmp(token,"oscserver"))
			{
			char server[32];
			sscanf(line,"%s%s",token,server);
			config->useOSC = true;
			strcpy(config->oscServer,server);
			}
		else if (!strcmp(token,"oscport"))
			{
			int val;
			sscanf(line,"%s%d",token,&val);
			config->oscPort = val;
			}
		else if (!strcmp(token,"listenerfilter"))
			{
			float val;
			sscanf(line,"%s%f",token,&val);
			config->listenerFilter = val;
			}
		else if (!strcmp(token,"soundfilter"))
			{
			float val;
			sscanf(line,"%s%f",token,&val);
			config->soundFilter = val;
			}
		}
	if (curSpeakers > 0)
		config->numSpeakers = curSpeakers;
	fclose(fp);
	}

main(int argc,char **argv)
	{	
	printf("snerd (bergen sound server) - Version 0.6 - Compiled " __DATE__ "\n");
	ConfigInformation* config = (ConfigInformation*)malloc(sizeof(ConfigInformation));
	config->mode = NO_SPATIALIZATION;
	config->numSpeakers = 0;
	config->sampleRate=22050;
	config->sampleRecRate=22050;
	config->bufferSize=1024;
	config->udpPort=5900;
	config->useOSC=false;
	config->oscPort=7001;
	config->listenerFilter=0.2;
	config->soundFilter=0.2;
	char file[256] = {"/usr/local/bergen/etc/snerd.config"};
	loadConfigFile(config,file);
	char host[256];
	gethostname(host,256);
	sprintf(file,"/usr/local/bergen/etc/%s.config",host);
	loadConfigFile(config,file);
	strcpy(file,getenv("HOME"));
	strcat(file,"/.snerdrc");
	loadConfigFile(config,file);
	loadConfigFile(config,"./.snerdrc");
	switch(config->mode)
		{
		case STEREO:
			printf("Using Stereo spatialization with %d speakers\n",config->numSpeakers);
			config->outputChannels=2;
			config->inputChannels=1;
			config->sampleRecRate = config->sampleRate*2;
			break;
		case AMBISONIC:
			printf("Using Quad spatialization with %d speakers\n",config->numSpeakers);		
			config->outputChannels=4;
			config->inputChannels=4;
			config->sampleRecRate = config->sampleRate;
			break;
		default:
			printf("Using no spatialization with %d speakers\n",config->numSpeakers);		
			config->outputChannels=1;
			config->inputChannels=1;
			config->sampleRecRate = config->sampleRate;
			break;
		}
	if (config->useOSC)
		{
#ifdef OSC_DIR
		OSCServer::init(config);
		OSCServer::mainLoop();
#else
		printf("OSC library reguired for MaxMSP\n");
#endif
		}
	else
		{
		AudioInterface::openInputPort(config->bufferSize,config->sampleRecRate,config->inputChannels);
		AudioInterface::openOutputPort(config->bufferSize,config->sampleRate,config->outputChannels);
		snerdServer::init(config);
		snerdServer::mainLoop();
		}
	}
